package edu.berkeley.qfat;
import java.awt.*;
import java.io.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import java.util.*;
import edu.berkeley.qfat.geom.*;
import edu.berkeley.qfat.geom.Point;
import edu.berkeley.qfat.geom.Polygon;

// TO DO:
//
// - Implement "real" constraints (plane, line, point)
//
// - Constrained surface subdivision
//     - Edge.flip() gets us the triforce subdivision
//     - Edge.delete() gets us the catmull-clark subdivision
//     - Catmull-Clark: just don't move points if we can't.  Need to average the influence of the points on a binding group.
//
// - Ability to snap three views to orthgonal
// - SLIDE UI
//     - left button -> crystal ball
//     - translate
//     - rightbutton/mousewheel zoom
// - v+click to select vertex
// - show: constraints, bindings, faces
//
// Editing:
//  - fracture edge, face
//  - change aspect ratio
//  - ability to select a point, rotate the model, then move the point
//  - when moving a vertex in one window, show that window's axes in all other windows
//


// TO DO:
// - real anneal
// - solve self-intersection problem
// - get a better test model?
// - symmetry constraints withing the tile
// - rotation matrices
// - overbinding results in forced equational constraints on the leader
// - shatter in invertd-triforce pattern brian mentioned
// - aspect ratio?  non-uniform deformation?
// - rotational alignment

// - movie-style user interface like
//      http://www.coleran.com/markcoleranreell.html ?

// - consider recasting the Shewchuk predicates in Java?
//    http://www.cs.cmu.edu/afs/cs/project/quake/public/code/predicates.c

/*
  blender keys
  - middle mouse = option+click
  - right mouse = command+click

  3,7,1 = view along axes (control for opp direction)
  4, 8, 7, 2 = rotate in discrete increments (+control to translate)
  middle trag: rotate space
  shift+middle drag: translate space
  wheel: zoom
  home: home view: take current angle, zoom to whole scnee
  5 = ortho vs non-ortho

*/

/*


 */


// FIXME: re-orient goal (how?)

public class Main extends InteractiveMeshViewer {

    public Matrix[] transforms;

    public static int verts = 1;

    public static final Random random = new Random();
    
    /** magnification factor */
    private static final float MAG = 1;
    public static final float MATCHING_EPSILON = 0.001f;

    private static boolean small(float f) { return Math.abs(f) < 0.001; }
    public void generateTile(Matrix[] matrices, Mesh mesh) {
        mesh.coalesce = true;
        HashSet<HalfSpace> halfSpaces = new HashSet<HalfSpace>();
        HashSet<Polygon> polygons = new HashSet<Polygon>();
        for(Matrix m : matrices) {
                Vec v = m.getTranslationalComponent();
                if (v.mag() < 0.0001) continue;
                v = v.times(-1);
                v = v.times(0.5f);
                Point p = Point.ORIGIN.plus(v);
                v = v.times(-1);
                
                //System.out.println(v);
                HalfSpace hs = new HalfSpace(p, v.norm());
                halfSpaces.add(hs);
                polygons.add(new Polygon(hs));
        }
        for(Polygon p : polygons) {
            System.out.println(p.plane.norm + " " + p.plane.dvalue);
            for(HalfSpace hs : halfSpaces) {
                if (p.plane==hs) continue;
                p = p.intersect(hs);
            }
            p.tesselate(mesh);
        }
    }

    private void quad(Mesh mesh, Matrix m, Point p1_, Point p2_, Point p3_, Point p4_) {
        Point p1 = m.times(p1_);
        Point p2 = m.times(p2_);
        Point p3 = m.times(p3_);
        Point p4 = m.times(p4_);
        Point c  = new Point((p1.x+p2.x+p3.x+p4.x)/4,
                             (p1.y+p2.y+p3.y+p4.y)/4,
                             (p1.z+p2.z+p3.z+p4.z)/4);
        mesh.newT(p1, p2, c, null, 0);
        mesh.newT(p2, p3, c, null, 0);
        mesh.newT(p3, p4, c, null, 0);
        mesh.newT(p4, p1, c, null, 0);
    }

    public void loadGoal(String file) {
        try {
            StlFile stlf = new StlFile();
            InputStream res = this.getClass().getClassLoader().getResourceAsStream(file);
            stlf.readBinaryFile(file, res);
            setGoal(new Mesh(false));
            for(int i=0; i<stlf.coordArray.length; i+=3) {
                Point p0 = new Point(stlf.coordArray[i+0].x * MAG, stlf.coordArray[i+0].y * MAG, stlf.coordArray[i+0].z * MAG);
                Point p1 = new Point(stlf.coordArray[i+1].x * MAG, stlf.coordArray[i+1].y * MAG, stlf.coordArray[i+1].z * MAG);
                Point p2 = new Point(stlf.coordArray[i+2].x * MAG, stlf.coordArray[i+2].y * MAG, stlf.coordArray[i+2].z * MAG);
                Vec n    = new Vec(stlf.normArray[i/3].x * MAG, stlf.normArray[i/3].y  * MAG, stlf.normArray[i/3].z * MAG);
                Mesh.T t  = goal.newT(p0, p1, p2, n, 0);
            }
            float goal_width  = goal.diagonal().dot(new Vec(1, 0, 0));
            float goal_height = goal.diagonal().dot(new Vec(0, 1, 0));
            float goal_depth  = goal.diagonal().dot(new Vec(0, 0, 1));
            
        } catch (Exception e) { throw new RuntimeException(e);}
    }
    public void fixupGoal() { fixupGoal(true, true); }
    public void fixupGoal(boolean recenter, boolean lock) {
        // translate to match centroid
        if (recenter)
            goal.transform(Matrix.translate(tile.centroid().minus(goal.centroid())));
        if (lock) {
            goal.makeVerticesImmutable();
            tile.error_against = goal;
            goal.error_against = tile;
        }
    }

    public Main(JFrame f) { super(f); }

    public synchronized void fixupTile() { 
        for(Matrix m1 : transforms) {
            for(Matrix m2 : transforms) {
                if (m1==m2) continue;
                for(Mesh.T t1 : tile) {
                    for(Mesh.T t2 : tile) {

                        Matrix m = m1.inverse().times(m2);
                        Point t1v1 = m.times(t1.v1().p);
                        Point t1v2 = m.times(t1.v2().p);
                        Point t1v3 = m.times(t1.v3().p);
                        if (t1v1.distance(t2.v1().p) < MATCHING_EPSILON &&
                            t1v2.distance(t2.v3().p) < MATCHING_EPSILON &&
                            t1v3.distance(t2.v2().p) < MATCHING_EPSILON) {
                            t2.e3().bindEdge(t1.e1(), m);
                            t2.e2().bindEdge(t1.e2(), m);
                            t2.e1().bindEdge(t1.e3(), m);
                        }
                        if (t1v2.distance(t2.v1().p) < MATCHING_EPSILON &&
                            t1v3.distance(t2.v3().p) < MATCHING_EPSILON &&
                            t1v1.distance(t2.v2().p) < MATCHING_EPSILON) {
                            t2.e3().bindEdge(t1.e2(), m);
                            t2.e2().bindEdge(t1.e3(), m);
                            t2.e1().bindEdge(t1.e1(), m);
                        }
                        if (t1v3.distance(t2.v1().p) < MATCHING_EPSILON &&
                            t1v1.distance(t2.v3().p) < MATCHING_EPSILON &&
                            t1v2.distance(t2.v2().p) < MATCHING_EPSILON) {
                            t2.e3().bindEdge(t1.e3(), m);
                            t2.e2().bindEdge(t1.e1(), m);
                            t2.e1().bindEdge(t1.e2(), m);
                        }
                        if (t1v1.distance(t2.v1().p) < MATCHING_EPSILON &&
                            t1v2.distance(t2.v2().p) < MATCHING_EPSILON &&
                            t1v3.distance(t2.v3().p) < MATCHING_EPSILON) {
                            t2.e1().bindEdge(t1.e1().pair, m);
                            t2.e2().bindEdge(t1.e2().pair, m);
                            t2.e3().bindEdge(t1.e3().pair, m);
                        }
                        if (t1v2.distance(t2.v1().p) < MATCHING_EPSILON &&
                            t1v3.distance(t2.v2().p) < MATCHING_EPSILON &&
                            t1v1.distance(t2.v3().p) < MATCHING_EPSILON) {
                            t2.e2().bindEdge(t1.e1().pair, m);
                            t2.e3().bindEdge(t1.e2().pair, m);
                            t2.e1().bindEdge(t1.e3().pair, m);
                        }
                        if (t1v3.distance(t2.v1().p) < MATCHING_EPSILON &&
                            t1v1.distance(t2.v2().p) < MATCHING_EPSILON &&
                            t1v2.distance(t2.v3().p) < MATCHING_EPSILON) {
                            t2.e3().bindEdge(t1.e1().pair, m);
                            t2.e1().bindEdge(t1.e2().pair, m);
                            t2.e2().bindEdge(t1.e3().pair, m);
                        }
                    }
                }
            }
        }
        tile.rebindPoints();
        System.out.println("tile volume: " + tile.volume());
        System.out.println("goal volume: " + goal.volume());
        tile.error_against = goal;
        goal.error_against = tile;
        fixupGoal();
    }

    public void breakit() {
        int oldverts = verts;
        if (verts > 2000 && !force) return;
        force = false;
        System.out.println("doubling vertices.");
        PriorityQueue<Mesh.E> es = new PriorityQueue<Mesh.E>();
        for(Mesh.T t : tile) {
            es.add(t.e1());
            es.add(t.e2());
            es.add(t.e3());
            Thread.yield();
            repaint();
        }
        for(int i=0; i<Math.min(oldverts,50); i++) {
            Mesh.E e = es.poll();
            verts++;
            e.shatter();
            Thread.yield();
            repaint();
        }
        System.out.println("now have " + verts + " vertices; max is 2000");
        tile.rebindPoints();
    }

    public boolean rand(double temp, Mesh.Vertex p) {

        p.reComputeErrorAround();
        double tile_error = tile.error();
        double goal_error = goal.error();

        float max = p.averageEdgeLength()/5;
        Vec v = new Vec(random.nextFloat(), random.nextFloat(), random.nextFloat());
        v = v.norm().times((random.nextFloat() - 0.5f) * max);
        Matrix m = Matrix.translate(v);
        boolean good = p.move(v, false);
        if (!good) { return false; }

        double new_tile_error = tile.error();
        double new_goal_error = goal.error();
        double tile_delta = (new_tile_error - tile_error) / tile_error;
        double goal_delta = (new_goal_error - goal_error) / goal_error;
        double delta = tile_delta + goal_delta;
        double swapProbability = Math.exp((-1 * delta) / temp);

        //boolean doSwap = good && (tile_delta <= 0 && goal_delta <= 0);
        //boolean doSwap = good && (tile_delta + goal_delta <= 0);
        //if (temp < 0.000001) doSwap = good && (tile_delta <= 0 && goal_delta <= 0);
        boolean doSwap = good && (Math.random() < swapProbability);

        // always move uphill if possible -- potential idea
        if (tile_delta <= 0 && goal_delta <= 0 && good) doSwap = true;
        if (hillclimb)
            doSwap = tile_delta <= 0 && goal_delta <= 0 && good;

        if (doSwap) {
            tile_error = new_tile_error;
            goal_error = new_goal_error;
            hits++;
            p.goodp = p.p;
        } else {
            p.move(v.times(-1), true);
            misses++;
        }
        p.reComputeErrorAround();
        return true;
    }

    float hits = 0;
    float misses = 0;
    public void anneal() throws Exception {
        double hightemp = 1;
        temp = hightemp;
        double last = 10;
        boolean seek_upward = false;
        double acceptance = 1;
        while(true) {
            synchronized(this) {
                if (!anneal) { repaint(); Thread.sleep(10); continue; }

            double ratio = (hits+misses==0) ? 1 : (hits / (hits+misses));
            hits = 0;
            misses = 0;
            double gamma = 1;
            acceptance = (ratio+acceptance)/2;
            accepts = (int)(Math.ceil(ratio*100));
            temps = (int)(Math.ceil(temp*1000));
            vertss = tile.size();
            if (breaks > 0) {
                while (breaks>0) {
                    breaks--;
                    breakit();
                }
                seek_upward = true;
                continue;
            } else if (acceptance > 0.96) gamma = 0.1f;
            else if (acceptance > 0.9)    gamma = 0.2f;
            else if (acceptance > 0.8)    gamma = 0.3f;
            else if (acceptance > 0.6)    gamma = 0.4f;
            else if (acceptance > 0.3)    gamma = 0.8f;
            else if (acceptance > 0.15)   gamma = 0.94f;
            else if (acceptance > 0.10)   gamma = 0.98f;

            else if (acceptance < 0.01)   breaks++;

            if (seek_upward) {
                if (acceptance > 0.25) seek_upward = false;
                else gamma = 2-gamma;
            }

            if (anneal)
                temp = temp * gamma;


            HashSet<Mesh.Vertex> hs = new HashSet<Mesh.Vertex>();
            for(Mesh.Vertex p : tile.vertices()) hs.add(p);
            Mesh.Vertex[] pts = (Mesh.Vertex[])hs.toArray(new Mesh.Vertex[0]);

            int count = 0;
            long then = System.currentTimeMillis();
            for(int i=0; i<300; i++) {
                if (anneal) {
                    Mesh.Vertex v = pts[Math.abs(random.nextInt()) % pts.length];
                    if (breaks>0) break;
                    if (!rand(temp,v)) { i--; continue; }
                    v.recomputeFundamentalQuadricIfStale();
                    v.recomputeFundamentalQuadricIfNeighborChanged();
                    count++;
                }
                Thread.yield();
                repaint();
            }

            System.out.println("temp="+temp + " ratio="+(Math.ceil(acceptance*100)) + " " +
                               "points_per_second=" +
                               (count*1000)/((double)(System.currentTimeMillis()-then)));
            for(Mesh.Vertex p : goal.vertices()) {
                p.quadricStale = true;
                p.recomputeFundamentalQuadricIfNeighborChanged();
            }
            }
        }
    }


    public static void main(String[] s) throws Exception {
        JFrame f = new JFrame();
        f.setLayout(new BorderLayout());
        Main main = new Main(f);
        f.add(main, BorderLayout.CENTER);
        f.setJMenuBar(main.new MyMenuBar());
        f.pack();
        f.show();
        f.setSize(900, 900);
        f.doLayout();
        main.anneal();
    }

    public class MyMenuItem extends JMenuItem implements ActionListener {
        public MyMenuItem(String s) {
            super(s);
            this.addActionListener(this);
        }
        public void actionPerformed(ActionEvent event) {
            synchronized(Main.this) {
                    hit();
            }
        }
        public void hit() {}
    }

    public void hexBrick(boolean offset, boolean rotated) {
        setTile(new Mesh(false));
                float width  = (float)0.8;
                float depth  = (float)0.08;
                float height = (float)0.4;
                float rshift =   width/2;
                float lshift = -(width/2);
                float halfup = 0;
                Point ltf = new Point(lshift,  (depth/2),  (height/2));
                Point mtf = new Point( 0.0,    (depth/2),  (height/2));
                Point rtf = new Point(rshift,  (depth/2),  (height/2));
                Point lbf = new Point(lshift, -(depth/2),  (height/2));
                Point mbf = new Point( 0.0,   -(depth/2),  (height/2));
                Point rbf = new Point(rshift, -(depth/2),  (height/2));

                Point ltc = new Point(lshift,  (depth/2), 0);
                Point mtc = new Point( 0.0,    (depth/2), 0);
                Point rtc = new Point(rshift,  (depth/2), 0);
                Point lbc = new Point(lshift, -(depth/2), 0);
                Point mbc = new Point( 0.0,   -(depth/2), 0);
                Point rbc = new Point(rshift, -(depth/2), 0);

                Point ltn = new Point(lshift,  (depth/2), -(height/2));
                Point mtn = new Point( 0.0,    (depth/2), -(height/2));
                Point rtn = new Point(rshift,  (depth/2), -(height/2));
                Point lbn = new Point(lshift, -(depth/2), -(height/2));
                Point mbn = new Point( 0.0,   -(depth/2), -(height/2));
                Point rbn = new Point(rshift, -(depth/2), -(height/2));

        
                Point[] points = new Point[] {
                    ltf,
                    mtf,
                    rtf,
                    lbf,
                    mbf,
                    rbf,

                    ltc,
                    mtc,
                    rtc,
                    lbc,
                    mbc,
                    rbc,

                    ltn,
                    mtn,
                    rtn,
                    lbn,
                    mbn,
                    rbn
                };


                // top
                tile.newT(ltf, mtf, mtc, null, 1);
                tile.newT(mtc, ltc, ltf, null, 1);
                tile.newT(mtf, rtf, rtc, null, 1);
                tile.newT(rtc, mtc, mtf, null, 1);

                tile.newT(ltc, mtc, mtn, null, 1);
                tile.newT(mtn, ltn, ltc, null, 1);
                tile.newT(mtc, rtc, rtn, null, 1);
                tile.newT(rtn, mtn, mtc, null, 1);

                // bottom (swap normals)
                tile.newT(mbf, lbf, mbc, null, 2);
                tile.newT(lbc, mbc, lbf, null, 2);
                tile.newT(rbf, mbf, rbc, null, 2);
                tile.newT(mbc, rbc, mbf, null, 2);

                tile.newT(mbc, lbc, mbn, null, 2);
                tile.newT(lbn, mbn, lbc, null, 2);

                tile.newT(rbc, mbc, rbn, null, 2);
                tile.newT(mbn, rbn, mbc, null, 2);


                // left
                tile.newT(ltf, ltc, lbc, null, 3);
                tile.newT(lbc, lbf, ltf, null, 3);
                tile.newT(ltc, ltn, lbn, null, 3);
                tile.newT(lbn, lbc, ltc, null, 3);

                // right (swap normals)
                tile.newT(rtc, rtf, rbc, null, 4);
                tile.newT(rbf, rbc, rtf, null, 4);
                tile.newT(rtn, rtc, rbn, null, 4);
                tile.newT(rbc, rbn, rtc, null, 4);

                // front
                tile.newT(ltn, mtn, mbn, null, 5);
                tile.newT(ltn, mbn, lbn, null, 5);
                tile.newT(mtn, rtn, rbn, null, 5);
                tile.newT(mtn, rbn, mbn, null, 5);

                // back
                tile.newT(mtf, ltf, mbf, null, 6);
                tile.newT(mbf, ltf, lbf, null, 6);
                tile.newT(rtf, mtf, rbf, null, 6);
                tile.newT(rbf, mtf, mbf, null, 6);

                if (offset) {
                  transforms = new Matrix[] {
                      Matrix.translate(new Vec(lshift,      0,    height)),
                      Matrix.translate(new Vec(lshift,      0,   -height)),
                      Matrix.translate(new Vec(rshift,      0,    height)),
                      Matrix.translate(new Vec(rshift,      0,   -height)),
                      Matrix.translate(new Vec( width,      0,         0)),
                      Matrix.translate(new Vec(-width,      0,         0)),
                      Matrix.translate(new Vec(lshift,  depth, 0)),
                      Matrix.translate(new Vec(lshift, -depth, 0)),
                      Matrix.translate(new Vec(rshift,  depth, 0)),
                      Matrix.translate(new Vec(rshift, -depth, 0)),
                      Matrix.ONE,
                  };
                } else if (rotated) {
                    HashSet<Mesh.E> es = new HashSet<Mesh.E>();
                    for(Mesh.T t : tile) {
                        es.add(t.e1());
                        es.add(t.e2());
                        es.add(t.e3());
                    }
                    for(Mesh.E e : es) {
                        if (e.p1.p.x == e.p2.p.x && e.p1.p.y == e.p2.p.y) continue;
                        if (e.p1.p.z == e.p2.p.z && e.p1.p.y == e.p2.p.y) continue;
                        if (e.p1.p.x == e.p2.p.x && e.p1.p.z == e.p2.p.z) continue;
                        e.shatter();
                    }
                    transforms = new Matrix[] {
                        Matrix.translate(new Vec(0,   0,    height)).times(Matrix.rotate(new Vec(0, 0, 1), (float)Math.PI)),
                        Matrix.translate(new Vec(0,   0,   -height)).times(Matrix.rotate(new Vec(0, 0, 1), (float)Math.PI)),
                        Matrix.translate(new Vec(0,  depth, 0)),
                        Matrix.translate(new Vec(0, -depth, 0)),
                        Matrix.translate(new Vec( width,  0,    0)),
                        Matrix.translate(new Vec(-width,  0,    0)),
                      Matrix.ONE,
                  };
                } else {
                  transforms = new Matrix[] {
                      Matrix.translate(new Vec(lshift,      0,    height)),
                      Matrix.translate(new Vec(lshift,      0,   -height)),
                      Matrix.translate(new Vec(rshift,      0,    height)),
                      Matrix.translate(new Vec(rshift,      0,   -height)),
                      Matrix.translate(new Vec( width,      0,         0)),
                      Matrix.translate(new Vec(-width,      0,         0)),
                      Matrix.translate(new Vec(0,  depth, 0)),
                      Matrix.translate(new Vec(0, -depth, 0)),
                      Matrix.ONE,
                  };
                }
                
                fixupTile();
    }

    public class MyMenuBar extends JMenuBar {

        public MyMenuBar() {

            JMenu tileMenu = new JMenu("Tile");
            JMenu goalMenu = new JMenu("Goal");
            JMenu hideMenu = new JMenu("Actions");

            hideMenu.add(new MyMenuItem("Start Anneal") { public void hit() { anneal = true; }});
            hideMenu.add(new MyMenuItem("Stop Anneal") { public void hit() { anneal = false; }});
            hideMenu.add(new MyMenuItem("Reset to high temperature") { public void hit() { temp = 1; }});
            hideMenu.add(new MyMenuItem("Subdivide surface") { public void hit() { breaks++; }});
            hideMenu.add(new MyMenuItem("Show Goal") { public void hit() { goalon = true; }});
            hideMenu.add(new MyMenuItem("Hide Goal") { public void hit() { goalon = false; }});
            hideMenu.add(new MyMenuItem("Show All Neighbors") { public void hit() { neighbors = true; }});
            hideMenu.add(new MyMenuItem("Show One Neighbor Wireframe") { public void hit() { neighborsWireOne = true; }});
            hideMenu.add(new MyMenuItem("Show All Neighbors Wireframe") { public void hit() { neighborsWire = true; neighborsWireOne = false;}});
            hideMenu.add(new MyMenuItem("Hide Neighbors") { public void hit() { neighborsWire = false; neighborsWireOne = false; neighbors = false; }});

            goalMenu.add(new MyMenuItem("Fish with face") { public void hit() {
                loadGoal("face.stl");
                //goal.transform(Matrix.rotate(new Vec(0, 1, 0), (float)(Math.PI/2)));
                goal.transform(Matrix.rotate(new Vec(0, 0, 1), (float)(Math.PI/2)));
                //goal.transform(Matrix.scale(1, 2.2f, 1));
                float factor = (float)Math.pow(tile.volume() / goal.volume(), 1.0/3.0);
                //factor = factor * 0.8f;
                goal.transform(Matrix.scale(0.3f));
                //goal.transform(Matrix.rotate(new Vec(0,1,0), (float)(Math.PI*1/3)));
                goal.transform(Matrix.rotate(new Vec(1,0,0), (float)(Math.PI*1/3)));
                fixupGoal(true, false);
                //goal.transform(Matrix.translate(new Vec(0.145f, 0, 0)));
                fixupGoal(false, true);
            }});
            goalMenu.add(new MyMenuItem("Fish") { public void hit() {
                loadGoal("fish.stl");
                //goal.transform(Matrix.rotate(new Vec(0, 1, 0), (float)(Math.PI/2)));
                goal.transform(Matrix.rotate(new Vec(0, 0, 1), (float)(Math.PI/2)));
                goal.transform(Matrix.scale(1, 2.2f, 1));
                float factor = (float)Math.pow(tile.volume() / goal.volume(), 1.0/3.0);
                factor = factor * 0.8f;
                goal.transform(Matrix.scale(factor));
                fixupGoal(true, false);
                //goal.transform(Matrix.translate(new Vec(0.145f, 0, 0)));
                fixupGoal(false, true);
            }});
            goalMenu.add(new MyMenuItem("Hammerhead Fish") { public void hit() {
                loadGoal("fish.stl");
                goal.transform(Matrix.rotate(new Vec(0, 1, 0), (float)(Math.PI/2)));
                goal.transform(Matrix.rotate(new Vec(0, 0, 1), (float)(3*Math.PI/2)));
                float factor = (float)Math.pow(tile.volume() / goal.volume(), 1.0/3.0);
                factor *= 0.75f;
                goal.transform(Matrix.scale(factor));
                fixupGoal();
            }});
            goalMenu.add(new MyMenuItem("Vertical Fish") { public void hit() {
                loadGoal("fish.stl");
                //goal.transform(Matrix.rotate(new Vec(0, 0, 1), (float)(Math.PI/2)));
                //goal.transform(Matrix.rotate(new Vec(1, 0, 0), (float)(Math.PI/2)));
                float factor = (float)Math.pow(tile.volume() / goal.volume(), 1.0/3.0);
                goal.transform(Matrix.scale(factor/1.6f));
                fixupGoal();
            }});
            goalMenu.add(new MyMenuItem("Torus") { public void hit() {
                loadGoal("torus.stl");
                float factor = (float)Math.pow(tile.volume() / goal.volume(), 1.0/3.0);
                goal.transform(Matrix.scale(factor/2.5f));
                goal.transform(Matrix.rotate(new Vec(0, 1, 0), (float)(Math.PI/2)));
                fixupGoal();
            }});
            tileMenu.add(new MyMenuItem("Hex Brick") { public void hit() {
                hexBrick(false, false);
            }});
            tileMenu.add(new MyMenuItem("Hex Brick, offset planes") { public void hit() {
                hexBrick(true, false);
            }});
            tileMenu.add(new MyMenuItem("Hex Brick, rotated") { public void hit() {
                hexBrick(false, true);
            }});
            tileMenu.add(new MyMenuItem("Temp (do not use)") { public void hit() {
                setTile(new Mesh(false));
                float width  = (float)0.8;
                float depth  = (float)0.08;
                float height = (float)0.4;

                float rshift =   width/2;
                float lshift = -(width/2);
                float halfup = 0;
                //float shift = height/2;
                //width = (width*2)/3;
                float shift = 0;
                transforms = new Matrix[] {

                    Matrix.translate(new Vec(lshift/2,  depth,    -shift)),
                    Matrix.translate(new Vec(rshift/2,  depth,    -shift)),
                    Matrix.translate(new Vec(lshift/2, -depth,    -shift)),
                    Matrix.translate(new Vec(rshift/2, -depth,    -shift)),

                    Matrix.translate(new Vec(lshift,  depth/2,    -shift)),
                    Matrix.translate(new Vec(rshift,  depth/2,    -shift)),
                    Matrix.translate(new Vec(lshift, -depth/2,    -shift)),
                    Matrix.translate(new Vec(rshift, -depth/2,    -shift)),


                    /*
                      Matrix.translate(new Vec(lshift,  depth,    -shift)),
                      Matrix.translate(new Vec(rshift,  depth,    -shift)),
                      Matrix.translate(new Vec(lshift, -depth,    -shift)),
                      Matrix.translate(new Vec(rshift, -depth,    -shift)),
                    */
                    /*
                      Matrix.translate(new Vec(lshift,  depth,    shift)),
                      Matrix.translate(new Vec(rshift,  depth,    shift)),
                      Matrix.translate(new Vec(lshift, -depth,    shift)),
                      Matrix.translate(new Vec(rshift, -depth,    shift)),
                    */
                    //Matrix.translate(new Vec(0,  depth,    0)).times(Matrix.rotate(new Vec(0, 0, 1), (float)Math.PI)),
                    //Matrix.translate(new Vec(0, -depth,    0)).times(Matrix.rotate(new Vec(0, 0, 1), (float)Math.PI)),
                    //Matrix.translate(new Vec(0,   0,    height)).times(Matrix.rotate(new Vec(0, 0, 1), (float)Math.PI)),
                    //Matrix.translate(new Vec(0,   0,   -height)).times(Matrix.rotate(new Vec(0, 0, 1), (float)Math.PI)),

                    //Matrix.translate(new Vec(0,  depth, 0)),
                    //Matrix.translate(new Vec(0, -depth, 0)),
                    Matrix.translate(new Vec(0,      0,  height)),
                    Matrix.translate(new Vec(0,      0, -height)),

                    //Matrix.translate(new Vec(lshift,        depth,  height/2)),
                    //Matrix.translate(new Vec(lshift,        depth, -height/2)),
                    //Matrix.translate(new Vec(rshift,       -depth,  height/2)),
                    //Matrix.translate(new Vec(rshift,       -depth, -height/2)),
                    //Matrix.translate(new Vec(rshift,       0,  height)),
                    //Matrix.translate(new Vec(rshift,       0, -height)),

                    Matrix.translate(new Vec( width,           0,    0)),
                    Matrix.translate(new Vec(-width,           0,    0)),

                    Matrix.ONE
                };
                fixupTile();
            } });
            tileMenu.add(new MyMenuItem("Dense Packing (hex)") { public void hit() {
                setTile(new Mesh(false));
                float width  = (float)3.2;
                float depth  = (float)0.32;
                float height = (float)1.6;
                float unit = 0.4f;
                float r = unit/2;
                float sin = (float)(unit * Math.sin(Math.PI/3));
                float cos = (float)(unit * Math.cos(Math.PI/3));
                float x = (float)(r*Math.tan(Math.PI/6));
                float z = (float)(r/Math.cos(Math.PI/6));
                height = 2*r*(float)Math.sqrt(2f/3f);

                /*
                r *= 0.3f;
                cos *= 0.3f;
                unit *= 0.3f;
                */

                /*
                  sin *= 0.3f;
                  x *= 0.3f;
                  z *= 0.3f;
                */
                transforms = new Matrix[] {
                    Matrix.translate(new Vec(-unit, 0, 0)),
                    Matrix.translate(new Vec( unit, 0, 0)),
                    Matrix.translate(new Vec(-cos,  0,  sin)),
                    Matrix.translate(new Vec( cos,  0,  sin)),
                    Matrix.translate(new Vec(-cos,  0, -sin)),
                    Matrix.translate(new Vec( cos,  0, -sin)),
                    Matrix.translate(new Vec( 0,  height, z)),
                    Matrix.translate(new Vec(-r,  height, -x)),
                    Matrix.translate(new Vec( r,  height, -x)),
                    Matrix.translate(new Vec( 0, -height, -z)),
                    Matrix.translate(new Vec(-r, -height, x)),
                    Matrix.translate(new Vec( r, -height, x)),
                    Matrix.ONE,
                };
                generateTile(transforms, tile);
                fixupTile();
            } });
            tileMenu.add(new MyMenuItem("Slim Dense Packing (Cubic)") { public void hit() {
                setTile(new Mesh(false));
                float unit = 0.4f;
                float r = unit/2;
                float sin = (float)(unit * Math.sin(Math.PI/3));
                float cos = (float)(unit * Math.cos(Math.PI/3));
                float x = (float)(r*Math.tan(Math.PI/6));
                float z = (float)(r/Math.cos(Math.PI/6));
                float height = 2*r*(float)Math.sqrt(2f/3f);

                transforms = new Matrix[] {
                    Matrix.translate(new Vec(-unit, 0, 0)),
                    Matrix.translate(new Vec( unit, 0, 0)),
                    Matrix.translate(new Vec(-cos,  0,  sin)),
                    Matrix.translate(new Vec( cos,  0,  sin)),
                    Matrix.translate(new Vec(-cos,  0, -sin)),
                    Matrix.translate(new Vec( cos,  0, -sin)),

                    /*
                    Matrix.translate(new Vec( 0,  height, -z)).times(Matrix.rotate(new Vec(0,1,0), (float)Math.PI)),
                    Matrix.translate(new Vec(-r,  height,  x)).times(Matrix.rotate(new Vec(0,1,0), (float)Math.PI)),
                    Matrix.translate(new Vec( r,  height,  x)).times(Matrix.rotate(new Vec(0,1,0), (float)Math.PI)),
                    Matrix.translate(new Vec( 0, -height, -z)).times(Matrix.rotate(new Vec(0,1,0), (float)Math.PI)),
                    Matrix.translate(new Vec(-r, -height,  x)).times(Matrix.rotate(new Vec(0,1,0), (float)Math.PI)),
                    Matrix.translate(new Vec( r, -height,  x)).times(Matrix.rotate(new Vec(0,1,0), (float)Math.PI)),
                    */

                    Matrix.translate(new Vec( 0,  height, -z)),
                    Matrix.translate(new Vec(-r,  height,  x)),
                    Matrix.translate(new Vec( r,  height,  x)),
                    Matrix.translate(new Vec( 0, -height,  z)),
                    Matrix.translate(new Vec(-r, -height, -x)),
                    Matrix.translate(new Vec( r, -height, -x)),
                    //Matrix.rotate(new Vec(0,0,1), (float)Math.PI),

                    /*
                    Matrix.translate(new Vec( 0,  height, -z)).times(Matrix.scale(-1,1,-1)),
                    Matrix.translate(new Vec(-r,  height,  x)).times(Matrix.scale(-1,1,-1)),
                    Matrix.translate(new Vec( r,  height,  x)).times(Matrix.scale(-1,1,-1)),
                    Matrix.translate(new Vec( 0, -height, -z)).times(Matrix.scale(-1,1,-1)),
                    Matrix.translate(new Vec(-r, -height,  x)).times(Matrix.scale(-1,1,-1)),
                    Matrix.translate(new Vec( r, -height,  x)).times(Matrix.scale(-1,1,-1)),
                    */
                    Matrix.ONE
                };
                generateTile(transforms, tile);
                //for(int i=0; i<transforms.length; i++) transforms[i] = Matrix.translate(m.times(vecs[i]));
                Matrix m = Matrix.scale(1.9f, 1f ,1);
                //Matrix m = Matrix.scale(1f, 2.1f, 1f);
                tile.transform(m);
                for(int i=0; i<transforms.length; i++)
                    transforms[i] = transforms[i].preMultiplyTranslationalComponentBy(m);
                fixupTile();

            } });
            tileMenu.add(new MyMenuItem("Genus-1") { public void hit() {
                synchronized(this) {
                    setTile(new Mesh(false));
                    Matrix mm = Matrix.scale(0.1f);
                    float height = 4;
                    float width  = 4;
                    float depth  = 1;
                    // top
                    quad(tile, mm, 
                         new Point( 2,  2,  0),
                         new Point( 1,  1, -1),
                         new Point(-1,  1, -1),
                         new Point(-2,  2,  0));
                    quad(tile, mm, 
                         new Point(-2,  2,  0),
                         new Point(-1,  1,  1),
                         new Point( 1,  1,  1),
                         new Point( 2,  2,  0));
                    quad(tile, mm, 
                         new Point( 1,  1, -1),
                         new Point( 1,  1,  1),
                         new Point(-1,  1,  1),
                         new Point(-1,  1, -1));

                    // bottom
                    quad(tile, mm, 
                         new Point(-2, -2,  0),
                         new Point(-1, -1, -1),
                         new Point( 1, -1, -1),
                         new Point( 2, -2,  0));
                    quad(tile, mm, 
                         new Point( 2, -2,  0),
                         new Point( 1, -1,  1),
                         new Point(-1, -1,  1),
                         new Point(-2, -2,  0));
                    quad(tile, mm, 
                         new Point(-1, -1, -1),
                         new Point(-1, -1,  1),
                         new Point( 1, -1,  1),
                         new Point( 1, -1, -1));

                    // left
                    quad(tile, mm, 
                         new Point( 2, -2,  0),
                         new Point( 1, -1, -1),
                         new Point( 1,  1, -1),
                         new Point( 2,  2,  0));
                    quad(tile, mm, 
                         new Point( 2,  2,  0),
                         new Point( 1,  1,  1),
                         new Point( 1, -1,  1),
                         new Point( 2, -2,  0));
                    quad(tile, mm, 
                         new Point( 1, -1, -1),
                         new Point( 1, -1,  1),
                         new Point( 1,  1,  1),
                         new Point( 1,  1, -1));

                    // bottom
                    quad(tile, mm, 
                         new Point(-2,  2,  0),
                         new Point(-1,  1, -1),
                         new Point(-1, -1, -1),
                         new Point(-2, -2,  0));
                    quad(tile, mm, 
                         new Point(-2, -2,  0),
                         new Point(-1, -1,  1),
                         new Point(-1,  1,  1),
                         new Point(-2,  2,  0));
                    quad(tile, mm, 
                         new Point(-1,  1, -1),
                         new Point(-1,  1,  1),
                         new Point(-1, -1,  1),
                         new Point(-1, -1, -1));

                    height = 4;
                    width  = 4;
                    depth  = 1;


                    transforms = new Matrix[] {
                        Matrix.translate(new Vec(0, 0.2f,0))
                        .times(Matrix.rotate(new Vec(0,1,0), (float)(1*Math.PI/2))),

                        Matrix.translate(new Vec(0,-0.2f,0))
                        .times(Matrix.rotate(new Vec(0,1,0), (float)(-1*Math.PI/2))),

                        Matrix.translate(new Vec( 0.2f,0,0))
                        .times(Matrix.rotate(new Vec(1,0,0), (float)(1*Math.PI/2))),

                        Matrix.translate(new Vec(-0.2f,0,0))
                        .times(Matrix.rotate(new Vec(1,0,0), (float)(-1*Math.PI/2))),


                        //Matrix.rotate(new Vec(0,0,1), (float)(1*Math.PI/2)),

                        /*
                          Matrix.rotate(new Vec(0,0,1), (float)(1*Math.PI/2)),

                          Matrix.rotate(new Vec(0,0,1), (float)(3*Math.PI/2)),
                          Matrix.rotate(new Vec(1,0,0), (float)(2*Math.PI/2)),
                        */

                        //Matrix.rotate(new Vec(0,0,1), (float)(2*Math.PI/2)),
                        //Matrix.scale(1,-1,1),

                        //Matrix.translate(new Vec( 0.2f, 0,0))
                        //.times(Matrix.rotate(new Vec(0,0,1), (float)( 1*Math.PI/2)))
                        //.times(Matrix.rotate(new Vec(0,1,0), (float)( 3*Math.PI/2))),

                        //Matrix.translate(new Vec(-0.2f, 0,0))
                        //.times(Matrix.rotate(new Vec(0,0,1), (float)( 3*Math.PI/2)))
                        //.times(Matrix.rotate(new Vec(0,1,0), (float)( 3*Math.PI/2))),
            
                        //Matrix.rotate(new Vec(0,0,1), (float)( 0*Math.PI/2))
                        //.times(Matrix.translate(new Vec(0, -0.2f, 0)))
                        //.times(Matrix.rotate(new Vec(0,1,0), (float)( 1*Math.PI/2))),
                        //Matrix.rotate(new Vec(0,0,1), (float)( 1*Math.PI/2))
                        //.times(Matrix.translate(new Vec(0, -0.2f, 0)))
                        //.times(Matrix.rotate(new Vec(0,1,0), (float)( 1*Math.PI/2))),

                        //Matrix.rotate(new Vec(0,0,1), (float)( 0*Math.PI/2))
                        //.times(Matrix.translate(new Vec(0, -0.2f, 0)))
                        //.times(Matrix.rotate(new Vec(0,1,0), (float)( 1*Math.PI/2))),
                        //Matrix.rotate(new Vec(0,0,1), (float)( 0*Math.PI/2))
                        //.times(Matrix.translate(new Vec(0, -0.2f, 0)))
                        //.times(Matrix.rotate(new Vec(0,1,0), (float)( 1*Math.PI/2))),

                        Matrix.ONE,
                    };
                    fixupTile();
                }}});
            tileMenu.add(new MyMenuItem("Hammerhead") { public void hit() {
                synchronized(this) {
                    setTile(new Mesh(false));
                    Matrix mm = Matrix.ONE;
                    float height1 = .1f;
                    float height2 = .1f;
                    float width  = .4f;
                    float depth  = .1f;
                    // top
                    Point a1 = new Point( -width/2,         height2/2, -depth/2);
                    Point b1 = new Point(        0,         height2/2, -depth/2);
                    Point c1 = new Point(        0, height1+height2/2, -depth/2);
                    Point d1 = new Point(  width/2, height1+height2/2, -depth/2);
                    Point e1 = new Point(  width/2,         height2/2, -depth/2);
                    Point f1 = new Point(  width/2,        -height2/2, -depth/2);
                    Point g1 = new Point(  width/2,-height1-height2/2, -depth/2);
                    Point h1 = new Point(        0,-height1-height2/2, -depth/2);
                    Point i1 = new Point(        0,        -height2/2, -depth/2);
                    Point j1 = new Point( -width/2,        -height2/2, -depth/2);
                    Point a2 = new Point( -width/2,         height2/2,  depth/2);
                    Point b2 = new Point(        0,         height2/2,  depth/2);
                    Point c2 = new Point(        0, height1+height2/2,  depth/2);
                    Point d2 = new Point(  width/2, height1+height2/2,  depth/2);
                    Point e2 = new Point(  width/2,         height2/2,  depth/2);
                    Point f2 = new Point(  width/2,        -height2/2,  depth/2);
                    Point g2 = new Point(  width/2,-height1-height2/2,  depth/2);
                    Point h2 = new Point(        0,-height1-height2/2,  depth/2);
                    Point i2 = new Point(        0,        -height2/2,  depth/2);
                    Point j2 = new Point( -width/2,        -height2/2,  depth/2);

                    quad(tile, mm, a1, b1, i1, j1);
                    quad(tile, mm, c1, d1, e1, b1);
                    quad(tile, mm, b1, e1, f1, i1);
                    quad(tile, mm, i1, f1, g1, h1);

                    quad(tile, mm, j2, i2, b2, a2);
                    quad(tile, mm, b2, e2, d2, c2);
                    quad(tile, mm, i2, f2, e2, b2);
                    quad(tile, mm, h2, g2, f2, i2);

                    quad(tile, mm, d1, d2, e2, e1);
                    quad(tile, mm, e1, e2, f2, f1);
                    quad(tile, mm, f1, f2, g2, g1);
                    quad(tile, mm, h1, g1, g2, h2);
                    quad(tile, mm, i2, i1, h1, h2);
                    quad(tile, mm, j1, i1, i2, j2);
                    quad(tile, mm, a2, a1, j1, j2);
                    quad(tile, mm, a1, a2, b2, b1);
                    quad(tile, mm, c2, c1, b1, b2);
                    quad(tile, mm, c1, c2, d2, d1);

                    transforms = new Matrix[] {

                        mm.times(Matrix.translate(new Vec(   width,     0, 0))),
                        mm.times(Matrix.translate(new Vec(  -width,     0, 0))),
                        mm.times(Matrix.translate(new Vec(-width/2, height1+height2, 0))),
                        mm.times(Matrix.translate(new Vec( width/2, height1+height2, 0))),
                        mm.times(Matrix.translate(new Vec(-width/2,-height1-height2, 0))),
                        mm.times(Matrix.translate(new Vec( width/2,-height1-height2, 0))),

                        mm.times(Matrix.translate(new Vec(   width/2,     0,  depth))),
                        mm.times(Matrix.translate(new Vec(  -width/2,     0,  depth))),
                        mm.times(Matrix.translate(new Vec(         0, height1+height2,  depth))),
                        mm.times(Matrix.translate(new Vec(         0,-height1-height2,  depth))),

                        mm.times(Matrix.translate(new Vec(   width/2,     0, -depth))),
                        mm.times(Matrix.translate(new Vec(  -width/2,     0, -depth))),
                        mm.times(Matrix.translate(new Vec(         0, height1+height2, -depth))),
                        mm.times(Matrix.translate(new Vec(         0,-height1-height2, -depth))),

                        Matrix.ONE
                    };
                    fixupTile();
                }}});

            // Finally, add all the menus to the menu bar.
            add(tileMenu);
            add(goalMenu);
            add(hideMenu);
        }

    }

}