package edu.berkeley.qfat;
import java.io.*;
import java.nio.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import com.sun.opengl.util.*;
import java.util.*;
import edu.berkeley.qfat.geom.*;
import edu.berkeley.qfat.geom.Point;

public class MeshViewer extends JPanel implements GLEventListener, MouseListener, MouseMotionListener, KeyListener, MouseWheelListener {

    private float tz = 0;
    private float anglex = 0;
    private float angley = 0;

    private Mesh.Vertex closest = null;
    private Point       closestOriginallyAt = null;
    private int         mousex;
    private int         mousey;
    private Matrix      projection = null;
    private Point       clickPoint = null;
    private GLCanvas    glcanvas;
    private boolean     updateVisibilities = false;
    private boolean     mouseInside = false;

    private HashSet<Mesh> meshes = new HashSet<Mesh>();

    public synchronized void addMesh(Mesh m) { meshes.add(m); }
    public synchronized void removeMesh(Mesh m) { meshes.remove(m); }

    public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) { }
    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) { }

    public void addKeyListener(KeyListener kl) { glcanvas.addKeyListener(kl); }

    public synchronized void init(GLAutoDrawable gld) {
        GL gl = glcanvas.getGL();//gld.getGL();
        gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        gl.glViewport(0, 0, 500, 300);
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glClearDepth(1.0);
        gl.glDepthFunc(GL.GL_LEQUAL);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glMatrixMode(GL.GL_MODELVIEW);

        float mat_specular[] = { 0.5f, 0.5f, 0.5f, 0.5f };
        float mat_shininess[] = { 50.0f };
        gl.glShadeModel(GL.GL_SMOOTH);
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_DIFFUSE, mat_specular, 0);
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, mat_specular, 0);  
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT, new float[] { 0.3f, 0.3f, 0.3f, 0.3f }, 0);  
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, mat_shininess, 0);
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, new float[] { 1.0f,    4.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, new float[] { -10.0f, 10.0f,   10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT2, GL.GL_POSITION, new float[] { 10.0f, -10.0f,   10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT3, GL.GL_POSITION, new float[] { 10.0f,  10.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT4, GL.GL_POSITION, new float[] { -10.0f, 10.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT5, GL.GL_POSITION, new float[] { 10.0f, -10.0f,  -10.0f, 0.0f }, 0);
        gl.glEnable(GL.GL_LIGHTING);
        gl.glEnable(GL.GL_LIGHT0);

        gl.glColorMaterial(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE);
        gl.glEnable(GL.GL_COLOR_MATERIAL);

        display(gld);

        // hack to get around Mac OS bug
        IntBuffer buf = ByteBuffer.allocateDirect(9*4*4).order(ByteOrder.nativeOrder()).asIntBuffer();
        gl.glReadPixels(0,0, 1, 1, gl.GL_RGB, gl.GL_UNSIGNED_BYTE, buf);
    }

    public synchronized final void display(GLAutoDrawable drawable) {
        glcanvas.setSize(glcanvas.getParent().getWidth(), glcanvas.getParent().getHeight());

        GL gl = glcanvas.getGL();//drawable.getGL();
        GLU glu = new GLU();

        if (!mouseInside) gl.glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
        else gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        gl.glPointSize(5.0f);
        gl.glLoadIdentity();
        glu.gluPerspective(50, ((float)drawable.getWidth())/drawable.getHeight(), 0.5, 10);

        glu.gluLookAt(0, 0, -((tz/10)-1), 0, 0, 0, 0, 1, 0);
        gl.glRotatef(anglex/3, 0, 1, 0);
        gl.glRotatef(-(angley/3), 1, 0, 0);

        gl.glEnable(GL.GL_LIGHTING);
        gl.glShadeModel(GL.GL_SMOOTH);
        for(Mesh mesh : meshes)
            mesh.render(gl, Matrix.ONE);

        // highlight the point closest to the mouse; we do this here to avoid flicker
        if (closest != null) {
            gl.glDisable(GL.GL_LIGHTING);
            gl.glShadeModel(GL.GL_FLAT);
            gl.glColor3f(1,1,1);
            gl.glBegin(gl.GL_POINTS);
            closest.getPoint().glVertex(gl);
            gl.glEnd();
        }

        projection = Matrix.getProjectionMatrix(gl);

        if (updateVisibilities) {
            updateVisibilities = false;
            // update vertex visibilities
            updateVisibility(gl);
            
            //Matrix projection = Matrix.getProjectionMatrix(gl);
            double dist = Double.MAX_VALUE;
            closest = null;
            closestOriginallyAt = null;
            for(Mesh mesh : meshes)
                if (mesh.option_selectable)
                    for(Mesh.Vertex v : mesh.vertices()) {
                        if (!v.visible) continue;
                        Point p = projection.times(v.getPoint());
                        int x = (int)p.x;
                        int y = (int)p.y;
                        int mousex = (int)getMouse().x;
                        int mousey = (int)getMouse().y;
                        if (closest==null || (x-mousex)*(x-mousex)+(y-mousey)*(y-mousey) < dist) {
                            dist = (x-mousex)*(x-mousex)+(y-mousey)*(y-mousey);
                            closest = v;
                        }
                    }
        }
    }

    protected synchronized void updateVisibility(GL gl) {
        IntBuffer buf = ByteBuffer.allocateDirect(9*4*4).order(ByteOrder.nativeOrder()).asIntBuffer();
        gl.glFlush();
        gl.glDrawBuffer(GL.GL_BACK);
        gl.glReadBuffer(GL.GL_BACK);
        gl.glPixelStorei(GL.GL_PACK_ALIGNMENT, 1);
        gl.glFlush();
        gl.glDisable(GL.GL_LIGHTING);
        gl.glShadeModel(GL.GL_FLAT);
        gl.glColor3f(0,0,0);
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        for(Mesh mesh : meshes) mesh.render(gl, Matrix.ONE);
        for(Mesh mesh : meshes)
            if (mesh.option_selectable)
                for(Mesh.Vertex v : mesh.vertices()) {
                    Point p = v.getPoint();
                    gl.glColor3f(1,1,1);
                    gl.glBegin(gl.GL_POINTS);
                    p.glVertex(gl);
                    gl.glEnd();
                    gl.glFlush();
                    
                    Point projected = projection.times(p);
                    gl.glReadPixels((int)projected.x-1, (int)projected.y-1, 3, 3, gl.GL_RGB, gl.GL_UNSIGNED_BYTE, buf);
                    
                    boolean vis = false;
                    for(int j=0; j<9*4; j++) vis |= buf.get(j)!=0;
                    v.visible = vis;
                    if (vis) {
                        gl.glColor3f(0,0,0);
                        gl.glBegin(gl.GL_POINTS);
                        p.glVertex(gl);
                        gl.glEnd();
                    }
                }
        gl.glShadeModel(GL.GL_SMOOTH);
        gl.glEnable(GL.GL_LIGHTING);
        gl.glDrawBuffer(GL.GL_FRONT);
    }

    /** return the position of the mouse as a point in window-space */
    public Point getMouse() {
        return new Point(mousex, glcanvas.getHeight()-mousey, 0);
    }

    /** return the position where the mouse button was pressed, or null if it is not currently pressed */
    public Point getMouseClick() {
        return clickPoint;
    }

    public void mouseWheelMoved(MouseWheelEvent e) {
        tz -= e.getWheelRotation();
    }

    public void keyTyped(KeyEvent e)  { }
    public void keyPressed(KeyEvent e)  { }
    public void keyReleased(KeyEvent e) { }

    public void mouseClicked(MouseEvent e) { }
    public void mouseEntered(MouseEvent e) { mouseInside = true; }
    public void mouseExited(MouseEvent e) { mouseInside = false; }
    public void mousePressed(MouseEvent e) {
        clickPoint = getMouse();
    }

    public void mouseReleased(MouseEvent e) {
        clickPoint = null;
    }

    public void mouseMoved(MouseEvent e) {
        mousex = e.getX();
        mousey = e.getY();
        if ((e.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) != 0)
            updateVisibilities = true;
    }

    public void mouseDragged(MouseEvent e) {
        if ((e.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) != 0) {
            if (closest != null && projection != null) {
                synchronized(this) {
                    if (closestOriginallyAt==null) closestOriginallyAt = closest.getPoint();
                    Vec d1 = projection.inverse().times(getMouse()).minus(projection.inverse().times(clickPoint));
                    Vec delta = d1.plus(closestOriginallyAt).minus(closest.getPoint());
                    closest.move(delta, false);
                }
            }
        } else {
            anglex -= mousex - e.getX();
            angley += mousey - e.getY();
        }
        mousex = e.getX();
        mousey = e.getY();
    }

    public MeshViewer() {
        glcanvas = new GLCanvas();
        glcanvas.addGLEventListener(this);
        setLayout(new BorderLayout());
        this.add(glcanvas, BorderLayout.CENTER);
        glcanvas.addMouseListener(this);
        glcanvas.addMouseMotionListener(this);
        glcanvas.addMouseWheelListener(this);
        glcanvas.addKeyListener(this);
    }

    public void repaint() {
        if (glcanvas != null) glcanvas.repaint();
    }

}